{% extends 'admin/model/create.html' %}
{% import 'bukuserver/lib.html' as buku with context %}

{% block tail %}
  {{ super() }}
  {{ buku.set_lang() }}
  {{ buku.limit_navigation_if_popup() }}
  {{ buku.brand_dbname() }}
  {{ buku.script('bookmark.js') }}
  {{ buku.fetch_checkbox(form.fetch.data) }}
  {{ buku.focus() }}
  {{ buku.link_saved() }}
{% endblock %}
