{% extends 'admin/model/modals/create.html' %}
{% import 'bukuserver/lib.html' as buku with context %}

{% block create_form %}
  {{ lib.render_form(form, return_url, lib.extra(), form_opts=form_opts,
                     action=url_for('.create_view', url=return_url),
                     is_modal=True) }}
{% endblock %}

{% block tail %}
  {{ super() }}
  {{ buku.script('bookmark.js') }}
  {{ buku.fetch_checkbox(form.fetch.data) }}
  {{ buku.focus('.modal-body') }}
{% endblock %}
